#
# $Id: Daemon.pm
#
# Client and server definition for an init script daemon to ensure
# init script runs don't overlap and end up killing off a service
# when trying to restart/reload the service many times during the same
# transaction.
#
# Designed with httpd in mind, but could work for any service provided
# the init script supports the status target.
# 

package Sauce::Service::Daemon;

$DEBUG = "1";
if ($DEBUG) {
    use Sys::Syslog qw( :DEFAULT setlogsock);
}

use lib qw(/usr/sausalito/perl);
use POSIX qw(setsid);
use File::Path qw(make_path);
use File::Basename;

my $QUEUE_DIR = '/usr/sausalito/services/';

# Frequency with which we check the queue:
my $QUEUE_CHECK_INTERVAL = 15;

# How long we wait before a service will be restarted:
my $QUEUE_WAIT_TIME = 10; 

sub new {
    my ($class) = @_;
    my $self = {};
    bless $self, $class;
    return $self;
}

sub _daemonize {
    my ($self) = @_;
    my $pid = fork();
    if (!defined($pid)) {
        die("Fork failed: $!\n");
    } elsif ($pid != 0) {
        exit(0);
    }
    setsid();
    $0 = 'sauce_serviced';
    close(STDIN);
    close(STDERR);
    close(STDOUT);
    open(STDOUT, '>', '/dev/null');
}

sub process_queue {

    opendir(my $dh, $QUEUE_DIR) || do {
        &_logmsg("Can't open $QUEUE_DIR: $!");
        die "Can't open $QUEUE_DIR: $!";
    };

    # &_logmsg("Processing service events in $QUEUE_DIR");

    while (my $file = readdir($dh)) {
        next if $file =~ /^\./;
        my $path = "$QUEUE_DIR$file";

        #&_logmsg("Processing service event file: $path");

        open(my $fh, '<', $path) || do {
            &_logmsg("Can't open $path: $!");
            next;
        };

        my $event = <$fh>;
        my $timestamp = <$fh>;

        close($fh);

        chomp($event);
        chomp($timestamp);

        my $current_time = time();
        my $time_difference = $current_time - $timestamp;

        # Stop gets executed (more or less) right away. Anything else must wait for $QUEUE_WAIT_TIME seconds: 
        if ($event eq 'stop' || (($event eq 'reload' || $event eq 'restart' || $event eq 'condrestart' || $event eq 'condreload') && $time_difference >= $QUEUE_WAIT_TIME)) {
            my $service = basename($file);
            my $system_call = "/usr/bin/systemctl $event $service.service";
            my $exit_code = system($system_call);

            if ($exit_code == 0) {
                &_logmsg("$system_call: Transaction successful");
                unlink($path);
            }
            else {
                &_logmsg("$system_call: Transaction failed with exit code $exit_code");
            }
        }
    }
    closedir($dh);
}

sub run {
    my ($self) = @_;
    $self->_daemonize();

    # Set up signal handlers
    $SIG{ALRM} = sub {
        local $!; # Clear errno to prevent "Interrupted system call" error
        process_queue();
        alarm($QUEUE_CHECK_INTERVAL);
    };

    $SIG{INT} = $SIG{TERM} = sub {
        terminate(1);
    };

    make_path($QUEUE_DIR) unless -d $QUEUE_DIR;
    alarm($QUEUE_CHECK_INTERVAL);

    &_logmsg("${0}[${$}] Ready to accept requests");

    while (1) {
        sleep($QUEUE_CHECK_INTERVAL);
    }
}


sub _logmsg {
    my $msg = shift;
    &debug_msg("$msg \n");
}

sub debug_msg {
    if ($DEBUG) {
        my $msg = shift;
        setlogsock('unix');
        openlog($0, '', 'user');
        syslog('info', "Daemon.pm: $msg");
        closelog;
    }
}

1;

# 
# Copyright (c) 2008-2023 Michael Stauber, SOLARSPEED.NET
# Copyright (c) 2008-2023 Team BlueOnyx, BLUEONYX.IT
# All Rights Reserved.
# 
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in 
#    the documentation and/or other materials provided with the 
#    distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its 
#    contributors may be used to endorse or promote products derived 
#    from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# 
# You acknowledge that this software is not designed or intended for 
# use in the design, construction, operation or maintenance of any 
# nuclear facility.
# 